
BufSeq {
		// intended for sequencing percussion parts using buffers to store patterns
	
	var	<target, <click, <outbus, <clock, <pattern, <thing,
		<defname, <buf, <sampbuf, <synth, <playArgs, <def, <patLength;
	
	*initClass {
		2.do({ arg i;
			SynthDef.writeOnce("BufSeqSample" ++ (i+1), { arg impbus = 0, seqbuf = 0, sampbuf,
					rate = 1, amp = 1, outbus = 0;
				var sig, gate, index, click, vel, scan;
				click = In.kr(impbus, 1);
				index = Stepper.kr(click, 0, 0, BufFrames.kr(seqbuf)-1, 1);
				gate = Trig.kr(Index.kr(seqbuf, index), SampleDur.ir);
				vel = Latch.kr(gate, gate);
				sig = PlayBuf.ar(i+1, sampbuf, BufRateScale.kr(sampbuf)*rate, gate, 0, 0)
					* amp * vel;
				Out.ar(outbus, sig);
			});
		});
	}
	
		// patLength = # of beats in loop
		// pattern = array of gates
	*new { arg target, click, patLength = 4, pattern, thing, args, outbus;
		^super.newCopyArgs(target, click, outbus, nil, pattern, thing)
			.init(patLength, args)
	}
	
	init { arg patLen, args;
		
		target.isKindOf(MixerChannel).if({
			outbus = target.inbus;
		}, {
			outbus = outbus ?? { Bus.new(\audio, 0, 1) };
		});
		target = target.asTarget;

		clock = click.tryPerform(\clock) ?? { 
			click.isKindOf(TempoClock).if(click, { TempoClock.default });
		};
		
		patLength = patLen;
		
//		pattern = this.getPattern(patLength, pattern);
		
		this.prepareForPlay(thing, args);
	}
	
	play { arg quant = 1, amp = 1;
		var startFunc, latency;

		startFunc = {
//["BufSeq-play", clock.elapsedBeats].postln;
			synth.notNil.if({ target.server.sendBundle(latency, ["/n_free", synth.nodeID]) });
			synth = Synth.basicNew(defname, target.server);
			target.server.sendBundle(latency, synth.addToTailMsg(target, playArgs ++
				[\impbus, click.index, \seqbuf, buf.bufnum, \amp, amp ? 1,
				 \outbus, outbus.index]))
		};
		
		defname.isNil.if({
"BufSeq-play: Not ready for play.\nSoundfile loading may have failed, or the server may not be running.".warn; ^this });
		
		(quant > 0).if({
			latency = (click.subdiv.reciprocal * 0.9);
			clock.schedAbs(quant.asTimeSpec.applyLatency(latency).schedTime(clock),
//			
//			clock.elapsedBeats.roundUp(quant) + offset
//				- (latency*clock.tempo),
				startFunc);
		}, {
			latency = 0;
			startFunc.value;
		});
	}
	
	stop { arg quant = 1;
		var latency;
		synth.notNil.if({
			(quant > 0).if({
				latency = click.subdiv.reciprocal * 0.9;
				clock.schedAbs(quant.asTimeSpec.applyLatency(latency).schedTime(clock) - 0.1, {
					synth.free;
					synth = nil;
				});
			}, {
				synth.free;
				synth = nil;
			});
		});
	}
	
	free { arg quant = 1;
		this.stop(quant);
		buf.free;
		sampbuf.free;
		defname = def = nil;
	}
	
	isPlaying { ^synth.notNil }
	
	readyForPlay { ^defname.notNil }
	
		// allow switching between different sequence buffers
	buf_ { arg b, quant = 1;
		var latency;
		synth.notNil.if({
			b.isKindOf(Buffer).if({ b = b.bufnum });
			(quant > 0).if({
				latency = click.subdiv.reciprocal * 0.9;
				clock.schedAbs(quant.asTimeSpec.applyLatency(latency).schedTime(clock) - 0.1, {
					target.server.sendBundle(latency, 
						["/n_set", synth.nodeID, \seqbuf, b])
				})
			}, {
				target.server.sendBundle(latency, 
					["/n_set", synth.nodeID, \seqbuf, b])
			});
		});
	}
	
		// note, this only changes THIS SEQUENCER'S PATTERN
		// if you switch to another buffer, this method will have no audible effect
		// until you switch back
	pattern_ { arg len, p, quant = 1;
		var latency, newpattern;
//["BufSeq-pattern_", len, p, quant, offset].postln;
		(p.size > buf.numFrames).if({
			"BufferSeq-pattern_: New pattern cannot be larger than old pattern.".die
		});
		newpattern = this.getPattern(len, p);
		this.readyForPlay.if({
			quant = quant.asTimeSpec;
			(quant.quant > 0).if({
				latency = click.subdiv.reciprocal * 0.9;
				clock.schedAbs(quant.applyLatency(latency).schedTime(clock) - 0.1, {
//["Sending new pattern", pattern, clock.elapsedBeats, latency].postln;
					pattern = newpattern;
					target.server.sendBundle(latency, 
						["/b_setn", buf.bufnum, 0, pattern.size] ++ pattern
//,[\n_set, debug.nodeID, \t_trig, 1]
					)
				})
			}, {
				pattern = newpattern;
				target.server.sendBundle(nil, 
					["/b_setn", buf.bufnum, 0, pattern.size] ++ pattern)
			});
		});
	}
	
	getPattern { arg len, p, selectbuf = false;	// assumes we already have a click
			// expands the pattern so there's an array entry for each subdivision
			// coming from the click. selectbuf is b/c the sample selector buffer
			// in MultiBufSeq has to be padded differently
		var	subdiv, ksub,		// subdivision needed, subdiv provided in click
			pad, padSize, pattern;
		
//["BufSeq-getPattern", len, p, subdiv, ksub].asCompileString.postln;
		subdiv = (p.size / len).asInteger * 2;	// 16 gates, 4 bts = subdiv of 4 (16th notes)
								// mul by 2 so successive gates can have 0's between them
								// if not integer division, gates at end will be thrown out
		
		click.isKindOf(TempoClick).if({ ksub = click.subdiv }, { ksub = 0 });

			// click's subdiv must be integer multiple of needed subdiv
		((ksub % subdiv != 0) || (subdiv < ksub)).if({
			subdiv = subdiv.lcm(ksub);	// get lowest common multiple
		});
			// if click's subdivision is not high enough, so make new
			// TempoClicks are cheap, using < 0.5% cpu, so no big deal
		(ksub < subdiv).if({
			click = TempoClick.new(target.server, clock, nil, subdiv);
		});

			// expand pattern, filling with 0's between supplied values
			// trigger pattern mus be padded with 0's, select pattern must duplicate each entry
		padSize = (subdiv / (p.size/len) - 1).asInteger;
		selectbuf.not.if({ pad = Array.fill(padSize, 0); });
		pattern = nil;
		
		(p.size - (p.size % len)).do({ arg i;
			pattern = pattern ++ selectbuf.if(
				{ Array.fill(padSize + 1, p.at(i)) },
				{ [p.at(i)] ++ pad });
		});
		
		^pattern//.postln
	}
	
	prepareForPlay { arg th, args;
			// if string, it's a sample name
			// if func, put it in a synthdef -- your func should have a gate argument
		var temppat;
		target.server.waitForBoot({
			th.isFunction.if({
				defname = "BufSeq" ++ Main.elapsedTime.trunc(0.01);
				def = SynthDef(defname, { arg impbus, seqbuf, amp = 1, outbus = 0;
					var sig, gate, index, click;
					click = In.kr(impbus, 1);
					index = PulseCount.kr(click, PulseDivider.kr(click, BufFrames.kr(seqbuf)));
					gate = BufRd.ar(1, seqbuf, index);
					sig = SynthDef.wrap(th, nil, [gate]) * amp;
					Out.ar(outbus, sig);
				}).send(target.server);
				playArgs = args;
			}, {
				sampbuf = Buffer.readAndQuery(target.server, th, 0, -1, {
						// use completion func to determine which synthdef to use
					arg b; 
					defname = "BufSeqSample" ++ b.numChannels;
				});
				playArgs = [\sampbuf, sampbuf.bufnum] ++ args;
			});
			
				// set up pattern
			temppat = this.getPattern(patLength, pattern);
			buf = Buffer.alloc(target.server, temppat.size, 1, { arg b;
				["/b_setn", b.bufnum, 0, temppat.size] ++ temppat
			});

		});
	}

	subdiv { ^click.subdiv }
	
}