
MultiBufSeq : BufSeq {
	// for sequencing mutually-exclusive perc. parts, like hi-hats
	// keeps an array of samples or functions and a second sequence
	// buffer keeps indices to choose which sample/func to play
	// since all ugens have to run at once, functions will be *real* inefficient here!
	
	// differences: sampbuf is an array of Buffers
	// thing is array of things (must all be same type--can't mix samples and funcs)
	
	
	var	<selectbuf,	// server buffer holding sample indices
		<selectpattern;
	
	*initClass {
		2.do({ arg i;
			SynthDef.writeOnce("BufSeqSampM" ++ (i+1), { arg impbus = 0, seqbuf = 0,
					selectbuf, rate = 1, amp = 1, outbus = 0;
				var sig, gate, index, click, vel, scan, sampbuf;
				click = In.kr(impbus, 1);
				index = Stepper.kr(click, 0, 0, BufFrames.kr(seqbuf)-1, 1);
				gate = Trig.kr(Index.kr(seqbuf, index), SampleDur.ir);
				sampbuf = Index.kr(selectbuf, index);  // which buf to play?
//SendTrig.kr(click, 0, sampbuf);
				vel = Latch.kr(gate, gate);
				sig = PlayBuf.ar(i+1, sampbuf, BufRateScale.kr(sampbuf) * rate, gate, 0, 0)
					* amp * vel;
				Out.ar(outbus, sig);
			});
		});
	}

	*new { arg target, click, patLength = 4, pattern, selectpattern, thing, args, outbus;
		^super.newCopyArgs(target, click, outbus, nil, pattern, thing)
			.init(patLength, args, selectpattern)
	}
	
	init { arg patLen, args, sp;
		
		target.isKindOf(MixerChannel).if({
			outbus = target.inbus;
		}, {
			outbus = outbus ?? { Bus.new(\audio, 0, 1) };
		});
		target = target.asTarget;

		clock = click.tryPerform(\clock) ?? { 
			click.isKindOf(TempoClock).if(click, { TempoClock.default });
		};
		
		patLength = patLen;
		selectpattern = sp;
		
		this.prepareForPlay(thing, args);
	}
	
	free { arg quant = 1;
		this.stop(quant);
		buf.free;
		sampbuf.do({ arg b; b.free; });
		defname = def = nil;
	}
	
	pattern_ { arg len, p, sp, quant = 1;
		var latency, newpattern, newselpat;
//["MultiBufSeq-pattern_", len, p, sp].asCompileString.postln;
		((p.size > (buf.numFrames/2)) || (sp.size > (buf.numFrames/2))).if({
			"BufferSeq-pattern_: New pattern cannot be larger than old pattern.".die
		});
		newpattern = this.getPattern(len, p);
		newselpat = this.getPattern(len, sp, true);
		newselpat = this.fixSelectPattern(newselpat);
		this.readyForPlay.if({
			quant = quant.asTimeSpec;
			(quant.quant > 0).if({
				latency = click.subdiv.reciprocal * 0.9;
				clock.schedAbs(quant.applyLatency(latency).schedTime(clock) - 0.1, {
					pattern = newpattern;
					selectpattern = newselpat;
					target.server.sendBundle(latency, 
						["/b_setn", buf.bufnum, 0, pattern.size] ++ pattern,
						["/b_setn", selectbuf.bufnum, 0, selectpattern.size] ++ selectpattern
//,[\n_set, debug.nodeID, \t_trig, 1]
					)
				})
			}, {
				pattern = newpattern;
				selectpattern = newselpat;
				target.server.sendBundle(nil, 
					["/b_setn", buf.bufnum, 0, pattern.size] ++ pattern,
					["/b_setn", selectbuf.bufnum, 0, selectpattern.size] ++ selectpattern
//,[\n_set, debug.nodeID, \t_trig, 1]
				)
			});
		});
	}

	prepareForPlay { arg th, args;
		var temppat;
		target.server.waitForBoot({
			th.isFunction.if({
				defname = "MBufSeq" ++ Main.elapsedTime.trunc(0.01);
				def = SynthDef(defname, { arg impbus, seqbuf, selectbuf, amp = 1, outbus = 0;
					var sig, gate, index, click, funcindex;
					click = In.kr(impbus, 1);
					index = PulseCount.kr(click, PulseDivider.kr(click, BufFrames.kr(seqbuf)));
					gate = BufRd.ar(1, seqbuf, index);
					funcindex = BufRd.ar(1, selectbuf, index);  // which buf to play?
					sig = SynthDef.wrap(th, nil, [gate]) * amp;
					Out.ar(outbus, sig);
				}).send(target.server);
				playArgs = args;
			}, {
					// this test allows you to pass in an array of pre-read buffers
				th[0].isKindOf(Buffer).not.if({
						// here, args is startFrame
					args.isKindOf(SimpleNumber).if({ args = [args, -1] });
					args.isNil.if({ args = [0, -1] });
					sampbuf = Array.new;
						// read all the files
					th.do({ arg t, i;
						sampbuf = sampbuf.add(
							Buffer.readAndQuery(target.server, t, args.at(0), args.at(1))
						);
					});
						// now wait for all files to be read
					Routine.new({
						{ BufferQueryQueue.isRunning }.while({ 0.5.wait });
							// now finish
							// can't do this until after the buffers are loaded
						("Finished loading " ++ th.size ++ " soundfiles.").postln;
						this.finishedLoading;
						nil.yield;	// stop
					}).play(SystemClock);
				}, {
					sampbuf = th;	// if you gave an array of buffers, use it
					this.finishedLoading;
				});
			});
			
				// set up pattern on server
			temppat = this.getPattern(patLength, pattern);
			buf = Buffer.alloc(target.server, temppat.size, 1, { arg b;
				["/b_setn", b.bufnum, 0, temppat.size] ++ temppat
			});
				// fix selectbuf to reflect actual buffer numbers of samples
		});
	}
	
	finishedLoading {
		var temppat;
		defname = "BufSeqSampM" ++ sampbuf.at(0).numChannels;
			// now set the selector pattern
		temppat = this.getPattern(patLength, selectpattern, true);
		temppat = this.fixSelectPattern(temppat);
		selectbuf = Buffer.alloc(target.server, temppat.size, 1, { arg b;
			["/b_setn", b.bufnum, 0, temppat.size] ++ temppat
		});
		playArgs = [\selectbuf, selectbuf.bufnum];
	}
	
	fixSelectPattern { arg sp;
		^sp.collect({ arg j;
			sampbuf.at(j).bufnum
		})
	}
	
}