
// todo: beatsPerBar in quantize defs
// translator classes are deprecated

MIDIRecBuf {
	var	<>name, <notes,	// notes is array of SequenceNotes
						// except, args will be velocity only
						// to play, must translate args into [\gate, Pseq(...)]
		<>absoluteOnsets = false,	// MIDIRecSocket now records as deltas,
								// not absolute times
		<>properties,		// a dict with user-defined characteristics
		
		>stopRecTime;		// a private variable to make sure last dur is correct
						// in convertToDeltas
		
	*new { arg name, notesArray, properties;
		^super.newCopyArgs(name, notesArray ?? { Array.new }).properties_(properties);
	}
	
	storeArgs { ^[name, notes, properties] }
	
	empty {
		notes = Array.new;
	}
	
	convertToDeltas {	// must be done only once!!
		var baseTime, runningTime, i;
		notes = notes.copy.sort({ |a, b| a.dur < b.dur });
		(absoluteOnsets and: (notes.size > 0)).if({
			baseTime = notes.first.dur;
			runningTime = 0;
			i = 0;
			{ i < (notes.size-1) }.while({
				notes[i].dur = notes[i+1].dur - runningTime - baseTime;
				runningTime = runningTime + notes[i].dur;
				i = i+1;
			});
			stopRecTime.notNil.if({
				notes[i].dur = stopRecTime - runningTime - baseTime;
				stopRecTime = nil;		// not needed anymore
			}, {
					// else, this wasn't a recorded buf and you really shouldn't be calling
					// this method. But, we have to fudge something.
				notes[i].dur = notes[i].length;
			});

			absoluteOnsets = false;
		});
	}
	
	convertGates {	// output a new Buf
		^this.class.new(name, notes.collect({ |n| 
			n.convertGates
		})).absoluteOnsets_(absoluteOnsets)
	}
	
	transpose { |steps = 0, newName|
		^(steps == 0).if({ this.copy.name_(newName ?? { name }) }, {
			newName.isNil.if({
				newName = (name ++ ("%%".format(
					steps.isPositive.if("+", ""),	// - will be inserted by rendering the #
					steps
				)));
			});
			this.class.new(newName, notes + steps, properties);
		});
	}
	
	dumpSeq {
		("\n\nMIDIRecBuf(" ++ 34.asAscii ++ name ++ 34.asAscii ++ ")").postln;
		notes.do({ |n, i| i.post; " : ".post; n.postln; })
	}
	
	asMIDIRecBuf { ^this }
	asPattern { ^Pseq(notes, 1) }
	asStream { ^this.asPattern.asStream }
	embedInStream { |inval| ^this.asPattern.embedInStream(inval) }
	
///////////// Some Collection-style methods for easier access
	
	at { arg i; ^notes.at(i) }	// also invoked by buf[index]
	
	first { ^notes.first }
	last { ^notes.last }
	
	add { arg note;
		notes = notes.add(note);
	}
	
	includes { arg item1;
		notes.do({ |item2| (item1 == item2).if({ ^true }); });
		^false
	}

		// useful for arpeggiator-processes
	removeDups { // arg key;	// key: maybe you want dup freqs removed, maybe gates, ...?
						// this is the method to get the value from the SequenceNote
						// SequenceNote supports only freq here
		var result;
//		(key == \gate).if({ key = \args });	// actually this won't work
		result = this.class.new(name, nil, properties);  // keep old name
		notes.do({ arg item;
			result.includes(item/*, key*/).not.if({ result.add(item) });
		});
		^result
	}
	
	sort { arg func;	// keys?
		^this.class.new(name, notes.copy.sort(func), properties);
	}
	
	++ { arg that;
		^this.class.new(name, notes ++ that.asArray, properties)
	}
	
	reverse { ^this.class.new(name, notes.reverse, properties) }
	
	scramble { ^this.class.new(name, notes.scramble, properties) }
	
	copyRange { arg start, stop;
		^this.class.new(name, notes.copyRange(start, stop).properties_(properties))
	}
	
	size { ^notes.size }
	
	asArray { ^notes }		// generally useful; see this-++ for a specific use
	
	copy { ^this.class.new(name, notes.deepCopy).properties_(properties) }
	
	copySeqNotes {
		^this.class.new(name, notes.collect(_.asSequenceNote)).properties_(properties)
	}
	
	copyMelNotes {
		^this.class.new(name, notes.collect(_.asMelodyNote)).properties_(properties)
	}
	
	copyTopNotes {
		^this.class.new(name, notes.collect(_.asTopNote)).properties_(properties)
	}
	
	mapMode { |mode|
		var	n;
		n = notes.collect({ |n| n.copy.freq_(n.freq.mapMode(mode)) });
		^this.class.new(name, n).properties_(properties)
	}
	
	unmapMode { |mode|
		var	n;
		n = notes.collect({ |n| n.copy.freq_(n.freq.unmapMode(mode)) });
		^this.class.new(name, n).properties_(properties)
	}
	
/////////////	These methods extract arrays from individual SequenceNotes,
///////////// for use in NoteSequences. You wrap the returned arrays in
///////////// whatever pattern you want.

	getArray { arg method = \freq;
		^notes.collect({ arg n; n.perform(method) });
	}

	midiNotes { ^this.getArray(\freq) }
	
	freqs {
		var	f = this.getArray(\freq),
			tuning = 	properties.tryPerform(\at, \tuning)
				?? { properties.tryPerform(\at, \mode).tryPerform(\asMode)
					.tryPerform(\cpsFunc) };
			
		tuning.notNil.if({ ^tuning.value(f) }, { ^f.midicps });
	}
	
	durs { ^this.getArray(\dur) }
	
	lengths { ^this.getArray(\length) }
	
	gates { ^this.getArray(\gate) }
	args { ^this.getArray(\args) }

		// deprecated	
//	translate { arg method = \simple;
//		^MIDISeqTranslator.translate(method, this);
//	}
	
	quantize { arg factor, beatsPerBar;
		var	onset = 0, lastOnset = 0,	// running onset times for each note
			newnote, prevnote, j;
		var b;
		factor = factor ?? { properties.tryPerform(\at, \factor) ?? { 0.25 } };
		beatsPerBar = beatsPerBar ?? { properties.tryPerform(\at, \beatsPerBar) };
		b = Array.new;
		notes.do({ arg note, i;
			b = b ++ (newnote = note.copy.dur_(note.dur.round(factor)))
				.length_(note.length * newnote.dur / note.dur);
			j = b.size - 2;	// scan backward for 0 length notes
			{ j >= 0 and: { b[j].length == 0 } }.while({
					// take original length and apply ratio to this note's (nonzero) dur
				b[j].length = notes[j].length * newnote.dur / notes[j].dur;
				j = j - 1;
			});
			onset = onset + newnote.dur;
//[note.length, newnote.dur, note.dur, note.length * newnote.dur / note.dur].postln;
		});
		beatsPerBar.notNil.if({
//"".postln; [lastOnset, b.collect({ |n| n.dur }).sum - b.last.dur, beatsPerBar, beatsPerBar - ((lastOnset - b.last.dur) % beatsPerBar)].postln;
				// fill out to end of bar if bar length specified
			b.last.dur_(beatsPerBar - ((onset - b.last.dur) % beatsPerBar));
		});
		^this.class.new(name, b, properties)

// deprecated
//			// factor = quantization granularity
//		^TranslatorDef.[\quantDeltaPreserveArtic].value(this, factor, beatsPerBar);
	}
	
		// attempt to quantize without requiring material to be exactly in tempo
		// that is, if tempo of buffer material is internally consistent but not
		// exactly in clock's tempo, try to find the tempo and adjust to the clock

		// this works well under some circumstances but fails in others -- unclear why

	baseRhythmicValue { arg factor, clock, error = 0.1;
		var	durs, /*error,*/ base, gcd,
			forwardAvg, backwardAvg, forwardSum, backwardSum, i;  // for outlier searching
		error = error * (clock ? TempoClock.default).tempo;  // at fast tempo, error must be higher
		
			// drop near-zero durs (these are chord notes that shouldn't figure into calcs)
			// sort them so shortest values come first
		durs = this.durs.select({ |item| item > error }).sort
				// clump durs into note values -- b-a is always >= 0 b/c array is sorted
			.separate({ |a, b| (b-a) > error });

//durs.do({ |dclump| [dclump.mean, dclump].asCompileString.postln });

			// base, by end, should be gcd of averages of all subarrays
			// if no gcd is found, base remains the average of lowest note value
		base = nil;
		durs.do({ |noteval|
			base.isNil.if({
				base = noteval.median;  // changing mean to median for now
			}, {
					// calculate running averages in both directions to search for outliers
					// this algorithm is more robust; however, if there's an outlier in
					// the smallest note value, it could still fail
				forwardAvg = Array.newClear(noteval.size);  // empty arrays for averages
				backwardAvg = Array.newClear(noteval.size);
				forwardSum = backwardSum = 0;
				noteval.size.reverseDo({ |i, j|
					forwardSum = forwardSum + noteval[j];
					backwardSum = backwardSum + noteval[i];
					forwardAvg.put(i, forwardSum / (j+1));
					backwardAvg.put(i, backwardSum / (j+1));
				});
				gcd = nil;	// test subsequent averages
				i = 0;
				{ gcd.isNil and: { i < forwardAvg.size }}.while({
					(gcd = base.fuzzygcd(forwardAvg[i], error)).notNil.if({
						base = gcd;
					}, {
						(gcd = base.fuzzygcd(backwardAvg[i], error)).notNil.if({
							base = gcd;
						});
					});
					i = i + 1;
				});		// end while
			});		// end base.isNil.if
//base.postln;
		});		// end durs.do
		
		(base < ((factor ? 0.25) / 2)).if({ ^nil }, { ^base });	// quantization failed, return nil
	}
	
	flexQuantize { arg factor, clock, error;
			// here, the algorithm will attempt to determine the base value present in the actual
			// notes played, and map that to the nearest "factor"
		var	buf, durs, base, adjustedBase, newdur;

		factor = factor ?? { properties.tryPerform(\at, \factor) ?? { 0.25 } };
		error = error ?? { properties.tryPerform(\at, \error) ?? { 0.1 } };
		
			// fail gracefully, or continue
		(base = this.baseRhythmicValue(factor, clock, error)).isNil.if({ ^nil });
		
		adjustedBase = base.round(factor);	// round base to nearest quant
//[base, adjustedBase].postln;
		buf = this.copy;	// make a new buf
		durs = this.durs;
		buf.notes.do({ |note, i|
				// scale to integer level, round to int, and scale back to adjusted value
			newdur = (durs[i] / base).round * adjustedBase;
//[durs[i], newdur].postln;
			(newdur > 0).if({
				note.length_(note.length * newdur / durs[i]);  // keep articulation as is
			}, {
				note.length_(note.length * adjustedBase / base);  // else scale note length
			});
			note.dur_(newdur);
		});
		
		^buf

			// MIDIRecBuf-parse supersedes this -- maybe reevaluate later
//		^TranslatorDef.[\consolidateChords].value(buf)
//			.name_(buf.name ++ ":q" ++ factor.round(0.001))  // identify as quantized
//			.properties_(properties)
//			.absoluteOnsets_(false)
	}
	
		// parse into more complex note types.
		// Allows sequences with chords and gracenotes to be quantized.
	parse { |deltaThresh, overlapThresh, allowShortNotes|
		var 	newbuf, notestemp, newNote, newNote2, oldIndex, startIndex, totalChordDelta,
			swaptemp;

			// cannot parse absolute onsets
		absoluteOnsets.if({
			^this.copy.convertToDeltas.parse(deltaThresh, overlapThresh, allowShortNotes);
		});

		deltaThresh = deltaThresh ?? { properties.tryPerform(\at, \deltaThresh) ?? { 0.1 } };
		overlapThresh = overlapThresh ?? { properties.tryPerform(\at, \overlapThresh) ?? { 0.1 } };
		allowShortNotes = allowShortNotes ??
			{ properties.tryPerform(\allowShortNotes) ?? { true } };

		newbuf = this.class.new(name ++ "b", []).properties_(properties);

		notestemp = notes;
		oldIndex = 0;
		{ oldIndex < notestemp.size }.while({
				// locate first note to be "main" (non-grace note)
			startIndex = oldIndex;
			{ oldIndex < notestemp.size and:
				{ notestemp[oldIndex].isGraceNote(deltaThresh, overlapThresh, allowShortNotes) }
			}.while({
					oldIndex = oldIndex + 1;
			});
				// create the note in the newbuf and add gracenotes
			newNote = notestemp[oldIndex] ?? { oldIndex = oldIndex - 1; notestemp.last };
			(oldIndex > startIndex).if({
				newNote = newNote.addGraceNotes(notestemp.copyRange(startIndex, oldIndex-1));
			});
				// while the overlap of previous note is high, they're chord notes
			totalChordDelta = notestemp[oldIndex].dur;
			oldIndex = oldIndex + 1;
			{ oldIndex < notestemp.size and:
					// calculates percentage of this note that overlaps with the last
					// if this note terminates before the main note, result > 1.0
					// which is OK b/c the test is >=
				{ notes[oldIndex-1].dur < deltaThresh and:
				  (((newNote.length - totalChordDelta) / notestemp[oldIndex].length)
				  		>= overlapThresh) } }.while({
						// got a chord note: now look for gracenotes within the chord
					startIndex = oldIndex;
					{ oldIndex < notestemp.size and:
						{ notestemp[oldIndex].isGraceNote(deltaThresh, overlapThresh) } }
					.while({
						oldIndex = oldIndex + 1;
					});
					newNote2 = notestemp[oldIndex];
					(oldIndex > startIndex).if({
						newNote2 = newNote2
							.addGraceNotes(notestemp.copyRange(startIndex, oldIndex-1));
					});
					newNote = newNote.add(newNote2);
					totalChordDelta = totalChordDelta + notestemp[oldIndex].dur;
					oldIndex = oldIndex + 1;
			});
				// done with this note
			newbuf.add(newNote);
		});
		(newbuf.notes.size > 0).if({
			^newbuf	// valid notes were created
		}, {
			^this		// all grace notes, just use the plain buffer
		});
	}

}

MIDIBufManager {
	// allows you to record midi note streams into multiple buffers (client-side)
	// controllers are ignored at present, maybe I'll add that later

		// when adding a buf, if I already have one of the same name, overwrite or throw error?
	classvar	<>overwrite = true;

	var	<active,	// am I on?
		<bufs,	// array of MIDIRecBufs
		<midiControl,	// MIDIController to choose a buf
		<midiChannel,
		<recorder,
		<clock,
		<value = 0,		// which buf am I looking at now?
		<>postRecFunc,	// executed after recording; user-definable, can set properties for buf
						// receives buf as arg, should return new buf
		<>view,		// will be guiable
		<>name;		// see asString
	
	*new { arg clock, chan, ccnum;
		^super.new.init(clock, chan, ccnum)
	}

	init { arg cl, chan, ccnum;
		midiChannel = chan.asChannelIndex;
		clock = cl ?? { TempoClock.default };
			// make a midi controller if asked, otherwise 
		chan.notNil.if({
			midiControl = BufManagerMIDIControl(midiChannel, ccnum, this);
		});
		bufs = List.new;
		active = true;
	}
	
	free {
		recorder.notNil.if({ recorder.stopRecord; });	// no need to keep going
		view.notNil.if({ view.remove });
		midiControl.free;
		bufs = midiControl = midiChannel = recorder = clock = value = view = nil;  // garbage
		active = false;
		MIDIPort.update;
	}
	
	add { arg buf;	// bufs can be added programmatically
		var index, nameTemp;
		bufs.includes(buf).not.if({
				// is there already a buf with this name?
			nameTemp = buf.name.asSymbol;
			(index = bufs.detectIndex({ |b| b.name.asSymbol == nameTemp })).notNil.if({
				overwrite.if({ bufs[index] = buf; },
					{ MethodError("MIDIRecSocket-add: already have a buf with name "
						++ nameTemp, this).throw });
			}, {
				bufs.add(buf);
			});
			view.notNil.if({ view.refresh })
		})
	}
	
	remove { arg buf;
		var res;
		(res = bufs.remove(buf)).notNil.if({
			view.notNil.if({ view.refresh });
		});
		^res
	}
	
	removeAt { |index|
		var res;
		(res = bufs.removeAt(index)).notNil.if({
			view.notNil.if({ view.refresh });
		});
		^res
	}
	
	removeCurrent {
		^this.removeAt(value)
	}
	
	value_ { arg v, updateGUI = true;
		value = v.wrap(0, bufs.size);	// ensure value is in array range
		(updateGUI and: { view.notNil }).if({ view.refresh(this); });
	}
	
	current { ^bufs.at(value) }
	
	at { |index|
		index.isInteger.if({ ^bufs[index] }, {
				// index may be buf name also -- index -1 if not found, returns nil
			^bufs[bufs.collect({ |buf| buf.name.asSymbol }).indexOf(index.asSymbol) ? -1]
		});
	}
	
	initRecord {
		recorder = MIDIRecSocket(midiChannel, clock);
		view.notNil.if({ view.refresh });
	}
	
	stopRecord {
		var newBuf;
		newBuf = recorder.stopRecord;
		protect {
			(newBuf.size > 0).if({
					  // if postRecFunc is nil, don't change value
				newBuf = postRecFunc.value(newBuf) ? newBuf;
	
// how do I know if postRecFunc set the name (to avoid resetting by user?)
	
				(value == bufs.size).if({	// creating a new buffer
					bufs.add(newBuf);
					view.tryPerform(\setName);	// focus namesetter field
				}, {
					bufs.put(value, newBuf);
					newBuf.name = bufs[value].name;	// updating old one, keep name
				});
			});
		} {		// must do this before throwing error
			recorder = nil;	// must garbage to be ready for next record call
			view.notNil.if({ view.refresh });
		};
			// if there was no error, the protect-unwind func doesn't get called,
			// so I must repeat here
		recorder = nil;	// must garbage to be ready for next record call
		view.notNil.if({ view.refresh });
	}
	
	guiClass { ^MIDIRecGUI }
	
	menuItems {
		^bufs.asArray.collect({ arg b; b.name }) ++ ["Create new"]
	}
	
//	draggedIntoVoicerGUI { arg dest;
//		var	v, seq;	// voicer to update
//		
////"MIDIBufManager-draggedIntoVoicerGUI".postln; this.dump; dest.dump;
//
//		v = dest.model;	// dest is a VoicerGUI; its model is a Voicer (or Ref to Voicer)
//			// dependence on view here means that this method should not be called
//			// unless a gui is present (which it always will be if you're dragging)
//		(value == bufs.size).if({	// can't translate "Create new"
//			^nil
//		});
//		seq = this.asNoteSequence;
//		seq.isNil.if({	// if seq is nil, something went wrong
//			"MIDIBufManager-draggedIntoVoicerGUI: invalid translator ("
//				++ view.currentTranslator ++ "). Transfer ignored.".warn;
//		}, {
//			seq = [bufs.at(value).name, { arg p, g; g.stopOthers(p); p.reset.play },
//				LinkedVoicerSequencer(v, seq)];
//			(v.processes.size == 0).if({	// must make a new process group
//				v.addProcess([
//					seq,
//					["Stop", { arg p, g; g.stopAll }]
//				]);
//					// auto-add to VPTrigSocket? what's current TrigSocket?
//			}, {
//					// how to determine current group? for now just use vpg 0
//				v.processes.at(0).add(seq);
//					// check that stop is there?
//			});
//		});
//		
//	}
	
	asMIDIRecBuf { ^this.current }
	
//	asNoteSequence {
//		^Library.at(\bufTrans, view.currentTranslator)
//			.tryPerform(\translate, bufs.at(value), inf);
//	}
	
	asString { ^"MidiBufManager : " ++ name }
	
}

// another issue: only one in a series of translators should convert to midicps
// document distinction between processors and convertors
// rept arg should be applied only to convertor

//MIDISeqTranslator {
//	// translates a MIDIRecBuf into a NoteSequence for use with VoicerSequencer
//	
//	classvar	<>addToLibrary = true;
//	
//	var	<>name,		// for gui menu
////		<>rept,		// default number of repeats in patterns for .translate
//		<>methodArgPairs;	// transformations and args
//	
//	*initClass {
//			// set up TranslatorDefs, which are automatically registered in the def class
//		Class.initClassTree(TranslatorDef);
//
//			// make default MIDISeqTranslator instances for the menu
//			// if you don't want, set addToLibrary to false
//			// putInLibrary may be set in the TranslatorDef for individual control
//		addToLibrary.if({
//			TranslatorDef.all.keysValuesDo({ arg key, xl;
//				xl.putInLibrary.if({
//					Library.put(\bufTrans, key,
//						MIDISeqTranslator.new(key, xl.name, xl.baseArgs));
//				});
//			});
//		});
//	}
//
//		// one-shot translation
//	
//	*translate { arg method, buf;
//		^Library.at(\bufTrans, method).translate(buf); // ?????
//	}
//
//		// create a translator object that can do a series of translations with one call
//	
//	*new { arg name ... pairs;	// \method, [args], \method, [args], etc.
//		var new;
////		pairs.at(0).isKindOf(Symbol).not.if({
////			pairs = pairs.flatten(1);	// pairs may have picked up an extra level of []
////		});
//		new = super.new.name_(name) /* .rept_(rept) */.methodArgPairs_(pairs.clump(2));
//		Library.put(\bufTrans, name.asSymbol, new);
//		^new
//	}
//	
//	translate { arg buf, rept;
//		var result, last;
//		result = buf;
////result.dumpSeq;
//		methodArgPairs.do({ arg pair;
//			result = TranslatorDef.at(pair.at(0)).value(result, pair.at(1));
////result.tryPerform(\dumpSeq);
//		});
//			// if you apply a processor only, the result won't be valid for play
//			// so this does the default conversion for the last xlator at the end if needed
//		result.isKindOf(NoteSequence).if({ ^result },
//			{ last = TranslatorDef.at(methodArgPairs.last.at(0));
//("executing final conversion: " ++ last.convertor).postln;
//			  ^TranslatorDef.at(last.convertor)
//				  .value(result, last.convertArgs)
//			});
//	}
//	
//	asEditString {	// print parameters formatted for easy editing
//		var out;
//		out = "MIDISeqTranslator( " ++ name.asCompileString ++ ",\n\n";
//		methodArgPairs.do({ arg pair, i;
//			out = out ++ "\t\t// "	// first, a comment to give arg names
//				++ TranslatorDef.at(pair[0]).listArgs ++ "\n";
//			out = out ++ "\t" ++ pair[0].asEditString ++ ", " ++ pair[1].asEditString
//				++ (i < (methodArgPairs.size-1)).if(",\n\n", "\n\n");
//		});
//		^out ++ ");\n\n" ++ TranslatorDef.listAll("// ");  // show available xlators w/args
//	}
//			
//}
//
//TranslatorDef {
//	classvar	<all;
//	var	<name, <func, <baseArgs, <convertor, <convertArgs,
//		<putInLibrary = true;
//				// true: a MIDISeqTranslator will be created for this TranslatorDef
//				// and placed in the library; you can override here
//
//// func should be: { arg buf, arg0, arg1; [create a new buf based on input, then give back result] }
//// baseArgs should be an array of default arguments that will be given to func
//// func is called with .valueArray([buf] ++ args)
//// see prototypes in *initClass
//// you may write your own TranslatorDefs well after the class tree is inited
//// they will be added to *all and available to use in MIDISeqTranslators (below)
//// -- this is all designed for maximum extensibility -- you can add things on the fly
//// without recompiling
//
//// TODO: how to set maxPlayBeats from gui
//
//	*new { arg name, func, baseArgs, convertor, convertArgs, putInLibrary = true;
//		var new;
//		new = super.newCopyArgs(name.asSymbol, func, baseArgs, convertor, convertArgs,
//			putInLibrary);
//		all.put(name, new);
//		^new
//	}
//	
//	*at { arg which; ^all.at(which) }
//	
//	value { arg buf ... args;
//		(args.size < baseArgs.size).if({		// not all args accounted for?
//			args = args ++ Array.newClear(baseArgs.size - args.size);
//		});
//		^func.valueArray([buf] ++ args.maskOver(baseArgs))
//	}
//	
//	*listAll { arg prefix = "";
//		^String.streamContents({ |stream|
//			all.asSortedArray.do({ |pair, i|
//				stream << prefix << pair[0] << " : ";
//				stream << (pair[1].listArgs ? "<no arguments>") << "\n";
//			});
//		});
//	}
//	
//	listArgs {
//		^String.streamContents({ |stream|
//			func.def.argNames.do({ |name, i|
//				(i > 1).if({ stream << ", " });
//				(i > 0).if({ stream << name });	// omit first arg, which is always buf
//			});
//		});
//	}
//	
//		// initClass must contain definitions of basic translators
//		// more may be added here by any user, or at runtime
//
//	*initClass {
////		Class.initClassTree(Collection);	// used to work, now it breaks Event.default
//		all = IdentityDictionary.new;
//
//		TranslatorDef(\simple, { arg buf, rept;	// convertor: makes the noteSequence
//			rept = rept ? 1;
//			NoteSequence(Pseq(buf.freqs, rept), Pseq(buf.durs, rept),
//				Pseq(buf.lengths, rept), true, nil,
//				[\gate, Pseq(buf.gates, rept)]);
//		}, [1]);		// baseargs: 1 repeat, no other args needed
//
//		TranslatorDef(\arpegConvert, { arg buf, rept,	// convertor for arpeggiation
//				durPat, lenPat, absLen, gatePat, otherArgPairs;
//			rept = rept ? inf;
//			NoteSequence(Pseq(buf.freqs, rept),
//				durPat ?? { Pseq(buf.durs, inf) }, 
//				lenPat ?? { Pseq(buf.lengths, inf) }, 
//				absLen ? false, nil,
//				[\gate, gatePat ?? { Pseq(buf.gates, inf) }] ++ otherArgPairs);
//		}, [inf, Pseq([0.25], inf), Pseq([1], inf), false, Pseq([1], inf), nil],
//			nil, nil, false);
//
//		TranslatorDef(\arpegUp, { arg buf;	// translator (uses \arpegConvert)
//// nil to use recorded info for pattern?
//			buf.removeDups.sort;	// assumes freq, ascending
//		}, [], \arpegConvert);
//
//		TranslatorDef(\arpegDown, { arg buf;	// translator
//// nil to use recorded info for pattern?
//			buf.removeDups.sort(\freq, \keyGT);
//		}, [], \arpegConvert);
//			
//		TranslatorDef(\arpegUpDown, { arg buf;
//			var b;
//			b = buf.removeDups.sort;
//			b ++ b.notes.reverse.copyRange(1, b.size-2);
//		}, [], \arpegConvert);
//		
//		TranslatorDef(\arpegDownUp, { arg buf;
//			var b;
//			b = buf.removeDups.sort;
//			b.reverse ++ b.notes.copyRange(1, b.size-2);
//		}, [], \arpegConvert);
//		
//		TranslatorDef(\arpegRandom, { arg buf;
//			buf.removeDups.scramble;
//		}, [], \arpegConvert);
//		
//			// like Pslide, but not using Pslide pattern
//			// because I don't want it to wrap around at the top
//		TranslatorDef(\arpegSlideUp, { arg buf, segmentLen, segmentStep, segmentStart;
//			var b;
//			buf = buf.removeDups.sort;
//			b = MIDIRecBuf(buf.name);	// empty
//			segmentStep = segmentStep.abs;	// can't be negative; use \arpegSlideDown for that
//			{ (segmentStart + segmentLen) < buf.size }.while({
//				b = b ++ buf.copyRange(segmentStart, segmentStart + segmentLen - 1);
//				segmentStart = segmentStart + segmentStep;
//			});
//			(segmentStart < buf.size).if({	// anything left over? grab it
//				b = b ++ buf.copyRange(segmentStart, buf.size-1);
//			});
//			b
//		}, [3, 1, 0], \arpegConvert);
//		
//		TranslatorDef(\reverse, { arg buf;
//			buf.reverse
//		}, \simple, [1]);
//		
//		TranslatorDef(\arpegSlideDown, { arg buf, segmentLen, segmentStep, segmentStart;
//			var b;
//				// so that 0 and index of list item have same effect
//			(segmentStart > (buf.size/2)).if({
//				segmentStart = buf.size - segmentStart - 1;
//			});
//			all.at(\arpegSlideUp).value(buf.reverse, [segmentLen, segmentStep, segmentStart]);
//		}, [3, 1, 0], \arpegConvert);
//		
//// this fails in new vm - with new SequenceNote structure this is not as robust as buf.parse anyway
//		TranslatorDef(\consolidateChords, {	// internal only
//			arg buf;
//			var res, i;
//			res = Array.with(buf.notes.first);	// start at the very beginning
//											// (a very good place to... oh, hell.)
//			i = 0;
//			{ i < (buf.size-1) }.while({	// loop to penultimate note
//				(buf[i].dur == 0).if({	// 0 dur, add next note to last chord in res
////					res.last.freq = res.last.freq.asArray ++ buf[i+1].freq;
////					res.last.length = res.last.length.asArray ++ buf[i+1].length;
//// VoicerSequencer does not now support different values for gate for diff notes in the same chord
////					res.last.args = res.last.args.asArray ++ buf[i+1].args;
//				}, {
//						// dur > 0, we can move on
////					res.last.dur = buf[i].dur;	// fill in this note value
////					res = res ++ buf[i+1];
//				});
//				i = i+1;
//			});
////			(res.last.dur == 0).if({	// make sure last note has a rhythmic value
////				res.last.dur = buf.notes.last.dur;
////			});
//			MIDIRecBuf(buf.name, res);
//		}, nil, nil, nil, false);
//
//		TranslatorDef(\quantOnset, { arg buf, quant, beatsPerBar;
//			var onset = 0, lastOnset = 0;	// running onset times for each note
//			var b;
//			b = Array.new;
//			buf.notes.do({ arg note, i;
//				onset = onset + note.dur;
//				b = b ++ note.copy.dur_(onset.round(quant) - lastOnset);
//				lastOnset = onset.round(quant);
//			});
//			beatsPerBar.notNil.if({
//					// fill out to end of bar if bar length specified
//				buf.last.dur_(beatsPerBar - (lastOnset % beatsPerBar));
//			});
//			MIDIRecBuf(buf.name, b);
//		}, [0.25, 4], \simple, [1]);
//		
//		TranslatorDef(\quantBoth, { arg buf, quant, beatsPerBar;
//			var onset = 0, lastOnset = 0;	// running onset times for each note
//			var b;
//			b = Array.new;
//			buf.notes.do({ arg note, i;
//				onset = onset + note.dur;
//				b = b ++ note.copy.dur_(onset.round(quant) - lastOnset)
//					.length_(note.length.round(quant));
//				lastOnset = onset.round(quant);
//			});
//			beatsPerBar.notNil.if({
//					// fill out to end of bar if bar length specified
//				buf.last.dur_(beatsPerBar - (lastOnset % beatsPerBar));
//			});
//			MIDIRecBuf(buf.name, b);
//		}, [0.25, 4], \simple, [1]);
//		
//		TranslatorDef(\quantPreserveArtic, { arg buf, quant, beatsPerBar;
//				// oldDur[i] - oldLength[i] == newDur[i] - newLength[i]
//			var	onset = 0, lastOnset = 0,	// running onset times for each note
//				newnote, prevnote, j;
//			var b;
//			b = Array.new;
//			buf.notes.do({ arg note, i;
//				onset = onset + note.dur;
//				b = b ++ (newnote = note.copy.dur_(onset.round(quant) - lastOnset))
//					.length_(note.length * newnote.dur / note.dur);
//				j = b.size - 2;	// scan backward for 0 length notes
//				{ j >= 0 and: { b[j].length == 0 } }.while({
//						// take original length and apply ratio to this note's (nonzero) dur
//					b[j].length = buf[j].length * newnote.dur / buf[j].dur;
//					j = j - 1;
//				});
////[note.length, newnote.dur, note.dur, note.length * newnote.dur / note.dur].postln;
//				lastOnset = onset.round(quant);
//			});
//			beatsPerBar.notNil.if({
////"".postln; [lastOnset, b.collect({ |n| n.dur }).sum - b.last.dur, beatsPerBar, beatsPerBar - ((lastOnset - b.last.dur) % beatsPerBar)].postln;
//					// fill out to end of bar if bar length specified
//				b.last.dur_(beatsPerBar - ((lastOnset - b.last.dur) % beatsPerBar));
//			});
////			all.at(\consolidateChords).value(MIDIRecBuf(buf.name, b));
//			MIDIRecBuf(buf.name, b)
//		}, [0.25, 4], \simple, [1]);
//
//		TranslatorDef(\quantDeltaPreserveArtic, { arg buf, quant, beatsPerBar;
//				// oldDur[i] - oldLength[i] == newDur[i] - newLength[i]
//			var	onset = 0, lastOnset = 0,	// running onset times for each note
//				newnote, prevnote, j;
//			var b;
//			b = Array.new;
//			buf.notes.do({ arg note, i;
////				(i > 0).if({ onset = onset + b.last.dur });
//				b = b ++ (newnote = note.copy.dur_(note.dur.round(quant)))
//					.length_(note.length * newnote.dur / note.dur);
//				j = b.size - 2;	// scan backward for 0 length notes
//				{ j >= 0 and: { b[j].length == 0 } }.while({
//						// take original length and apply ratio to this note's (nonzero) dur
//					b[j].length = buf[j].length * newnote.dur / buf[j].dur;
//					j = j - 1;
//				});
//				onset = onset + newnote.dur;
////[note.length, newnote.dur, note.dur, note.length * newnote.dur / note.dur].postln;
//			});
//			beatsPerBar.notNil.if({
////"".postln; [lastOnset, b.collect({ |n| n.dur }).sum - b.last.dur, beatsPerBar, beatsPerBar - ((lastOnset - b.last.dur) % beatsPerBar)].postln;
//					// fill out to end of bar if bar length specified
//				b.last.dur_(beatsPerBar - ((onset - b.last.dur) % beatsPerBar));
//			});
////			all.at(\consolidateChords).value(MIDIRecBuf(buf.name, b));
//			MIDIRecBuf(buf.name, b)
//		}, [0.25, nil], \simple, [1]);
//
//
//	}
//}
