
// miscellaneous Collection enhancements

+ SequenceableCollection {
		// find indices for every item found in another array
		// nils are ignored
	collectIndicesFromArray { arg sourceArray;
		var i, indices, me;
		me = this.flat;	// allow user to enable and disable many sockets
		indices = Array.new;
		me.do({ arg s;
			(i = sourceArray.indexOf(s)).notNil.if({
				indices = indices.add(i);
			});
		});
		^indices
	}
	
	collectIndicesOfItem { arg that;
		var	indices;
		indices = Array.new;
		this.do({ |item, i|
			(item === that).if({ indices = indices.add(i) })
		});
		^indices
	}
	
	collectIndices { |func|
		var	indices = Array(this.size);
		this.do({ |item, i|
			func.value(item, i).if({ indices = indices.add(i) });
		});
		^indices  // [0..indices.size-1] -- why did I write that? don't know what I was thinking
	}
	
	detectIndexFrom { |func, start = 0, step = 1|
		var	index = start;
		{	index.inclusivelyBetween(0, this.size-1)
			and: { func.(this[index]).not } }
		.while {
			index = index + step;
		};
		index.inclusivelyBetween(0, this.size-1).if { ^index } { ^nil }
	}
	
	detectFrom { |func, start = 0, step = 1|
		var	index;
		(index = this.detectIndexFrom(func, start, step)).notNil.if
			{ ^this[index] } { ^nil }
	}

	minIndex {
		^this.indexOf(this.minItem);
	}
	
	maxIndex {
		^this.indexOf(this.maxItem);
	}
	
	oneOfTop { arg num;
		^this.at(num.min(this.size).rand)
	}
	
	mapMode { arg mode;
		^this.collect({ |item| item.mapMode(mode) });
	}
	
	unmapMode { arg mode;
		^this.collect({ |item| item.unmapMode(mode) });
	}
	
		// change an array of notes into a SequenceNote with an array of frequencies etc.
	asChord {
		var	freqs, lengths, args;
		freqs = Array.new(this.size);
		lengths = Array.new(this.size);
		args = Array.new(this.size);
		this.do({ |note|
			freqs.add(note.freq);
			lengths.add(note.length);
			args.add(note.args);
		});
		^SequenceNote(freqs, this[0].dur, lengths, args)
	}
	
	avgsmooth { |avgpts = 5|
		var	runsum = this[..avgpts-1].sum,
			out = this.class.new(this.size - avgpts + 1);
		(this.size - avgpts + 1).do({ |i|
			out.add(runsum / avgpts);
			runsum = runsum - this[i];
			this[i+avgpts].notNil.if({ runsum = runsum + this[i+avgpts]; });
		});
		^out
	}		
}

+ SimpleNumber {
	mapMode { arg mode;
		^mode.prMap(this)	// save a dispatch by going right to the pseudo-private method
	}
	
	unmapMode { arg mode;
		^mode.prUnmap(this)
	}
}

+ Symbol {
		// no-op, just like \rest + 0
	mapMode {}
	unmapMode {}
}

// need a separate version of removeDups for Dictionary

+ Collection {
	removeDups {	// output a new collection without any duplicate values
		var result;
		result = this.species.new(this.size);
		this.do({ arg item;
			result.includes(item).not.if({ result.add(item) });
		});
		^result
	}
	
		// like -includes, but does not require the items to be identical
	includesEqual { arg x;
		this.do({ |item| (item == x).if({ ^true }); });
		^false
	}
	
//	median {
//		var temp;
//		temp = this.copy.sort;
//		temp.size.odd.if({
//			^temp[temp.size >> 1]	// bitshift to divide by 2 while preserving Integer class
//		}, {
//			^(temp[temp.size >> 1] + temp[(temp.size >> 1) - 1]) * 0.5;
//		});
//	}
//
//	asRational { 
//		^this.collect({ |item| item.asRational }) 
//	}
}

//+ SequenceableCollection {
//	median {
//		var temp;
//		temp = this.copy.sort;
//		temp.size.odd.if({
//			^temp[temp.size >> 1]	// bitshift to divide by 2 while preserving Integer class
//		}, {
//			^(temp[temp.size >> 1] + temp[(temp.size >> 1) - 1]) * 0.5;
//		});
//	}
//}

+ Dictionary {	// committed into main library now
	
//	collectAs { arg function, class;	// keep keys, function applies to value
//		var i, res;
//		res = class.new(this.size);
//		this.keysValuesDo({ arg k, v, i; res.put(k, function.value(v, i)) })
//		^res;
//	}
		// also committed
//	matchAt { |key|
//		this.keysValuesDo({ |k, v|
//			if(k.matchItem(key)) { ^v }
//		});
//		^nil
//	}
//
}

+ Array {
		// make a new copy of an array with items weighted as given in weights
		// weights should all be integers; non-integers will be rounded
	weight { arg weights;
		var res;
		weights = weights.round(1);
		res = Array.new;
		this.do({ arg item, i;
			res = res ++ Array.fill(weights.wrapAt(i) ? 1, item);
		});
		^res
	}
}

+ String {
	left { arg len; ^this.copyRange(0, len-1) }
	mid { arg start, len; ^this.copyRange(start, start + len - 1) }
	right { arg len; ^this.copyRange(this.size - len, this.size-1) }

	hexToInt {
		// assumes each character of string is a valid hex digit (0-9, A-F)
		// doesn't handle negatives
		var result = 0, i, size, digit;
		size = this.size - 1;
		i = size;		// start with rightmost digit
		{ (i >= 0) && result.notNil }.while({
			(digit = this.at(i).digit).notNil.if({
				result = result + (digit * (16 ** (size-i)));
			}, {
				result = nil;
			});
			i = i-1;
		});
		^result
	}
}

+ Symbol {
	hexToInt { ^this.asString.hexToInt }
}

+ Collection {
	hexToInt {
		^this.collect({ arg x; x.hexToInt })
	}
	
	asHexString {
		^this.collect({ arg x; x.asInteger.asHexString })
	}
		
}

+ Integer {
		// something like this was committed, so this is dropped
//	asHexString {
//		var str, isNeg, digit, num;
//		isNeg = this.isNegative;	// save for later
//		num = this.abs;			// decode as positive number
//		str = "";
//		{ num > 0 }.while({
//			digit = num & 15;		// get last 4 bits
//			str = (digit + (digit <= 9).if(48, 55)).asAscii.asString ++ str;  // prepend new digit
//			num = num >> 4;	// shift right for next 4 bits
//		});
//		(str.size == 0).if({ str = "0" });	// the while won't execute if this is 0
//		isNeg.if({ str = "-" ++ str });
//		^str
//	}
	
		// unrelated but I'm sticking it here anyway
		// return a string showing midi note
	asMIDINote { arg midcOct = 4;	// middle c = c4
		^(#["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]
			.at(this % 12))
			++ ((this / 12).asInteger - 5 + midcOct).asString
	}
}


+ SequenceableCollection {
		// lay this over that; where there are nils in this, substitute the
		// corresponding value in that
		// [0, nil, 2].maskOver([10, 1, 72]) --> [0, 1, 2]
	maskOver { arg that;
		^this.collect({ arg v, i; v ? that.wrapAt(i) });
	}

}

+ Nil {
	maskOver { arg that;
		^that
	}
}

+ Environment {
	*popAll {
		stack.size.do({ this.pop });
	}
}


+ Nil {
	nz { arg that = 1; ^that }	// so you don't have to nil-check before using nz
}

+ SimpleNumber {
	nz { arg that = 1; (this == 0).if({ ^that }); }	// this is returned by default

//	asRational { ^Rational.new(this) } 
//
//	/% { |that| ^Rational(this.asFloat / that) } 
}


// fuzzy gcd, needed for quantizing sequences

//+ Integer {
//	fuzzygcd { |that, error = 0.15, tries|
//			// error = acceptable amount of error in deciding "equality"
//			// tries = how many iterations before giving up & returning nil
//		^this.asFloat.fuzzygcd(that.asFloat, error, tries)
//	}
//}
//
//+ Float {
//	fuzzygcd { |that, error = 0.015, tries|
//		var lo, hi, margin;
//		lo = this.abs.min(that.abs);
//		hi = this.abs.max(that.abs);
//		tries = (tries ? 20).abs;
//		(((margin = (hi % lo) / lo) < error) or: (1-margin < error)).if({
//			(lo >= error).if({ ^lo }, { ^nil })	// if lo divides hi acceptably, return lo as gcd
//											// unless lo < error, then fail
//		}, {
//			(tries > 0).if({
//				^lo.fuzzygcd(hi % lo, error, tries - 1)  // recursive
//			}, {
//				^nil		// failed
//			});
//		});
//	}
//
//	/% { |that| ^Rational(this / that) }
//
//}

// maybe I'll write this later

//+ Collection {
//	cumulativeGCD {
//		var	temp, res;
//		temp = this.copy.sort.separate
//	}
//}

+ Ref {
	draggedIntoVoicerGUI { |dest| value.draggedIntoVoicerGUI(dest) }
}

// for use with SCStatusBox
// note that this will slow down regular posting behavior and is only experimental at this point

//+ String {
//	prPost { _PostString }
//	prPostln { _PostLine }
//	
//	post {
//		this.prPost;
//		SCStatusBox.post(this);
//	}
//	
//	postln {
//		this.prPostln;
//		SCStatusBox.postln(this);
//	}
//}

+ Symbol { isSymbol { ^true }}
+ Object {
	isSymbol { ^false }
	shallowCopyItems { ^this.shallowCopy }
	
	draggedIntoVoicerGCGUI { |gui|
		gui.model.spec = this.asSpec;
	}
}

	// currently same as copy or shallowCopy -- is that OK?
//+ Collection {
//	shallowCopyItems {
//		
//	}
//}

+ Dictionary {
		// shallowCopy copies everything except kernel objects like funcs
		// use this on each item of the dict -- do recursion on dictionaries within
	shallowCopyItems {
		var	new;
		new = this.class.new(this.size);
		this.keysValuesDo({ |k, v|
			new.put(k, v.shallowCopyItems);
		});
		^new
	}
}

+ IdentityDictionary {
	shallowCopyItems {
		^super.shallowCopyItems
			.proto_(proto)	// child object should not be modifying parent/proto
			.parent_(parent)	// so we don't copy them
			.know_(know)
	}
}

+ Spec {
	*listAll {
		specs.asSortedArray.do(_.postcs)
	}
		// Object implements this	
//	draggedIntoVoicerGCGUI { |gui|
//		gui.model.spec = this.asSpec;
//	}
}

+ Server {
	clumpListSendBundle { |time, bundle, delta = 0.05, clumpSize = 5|
		bundle.clump(clumpSize).do({ |segment, i|
			this.listSendBundle((time ? 0.1) + (i * delta), segment);
		});
	}
}

+ Event {
	*makeProto { |parentKey = \default|
		^Event.new(parent: parentEvents[parentKey])
	}
}

// flattening all collections (not just arrays)

+ Collection {
	asFlatArray { ^this.asArray.asFlatArray }
}

+ Object {
	asFlatArray { ^this }	// non-collections should remain as is - name is slightly misleading
}

+ SequenceableCollection {
	asFlatArray { ^this.collect(_.asFlatArray).flat }
}

// BroadcastServer compatibility

//+ BroadcastServer {
//	doesNotUnderstand { |selector ... args|
//		homeServer.perform(selector, *args)
//	}
//}

+ Object {
	releaseFromDependencies {
		dependantsDictionary.keysDo({ |obj|
			obj.dependants.includes(this).if({
				obj.removeDependant(this)
			});
		});
	}
}


// some general purpose default envelopes

+ Env {
	*zero { ^Env.new([0, 0], [1]) }
	
	*one { ^Env.new([1, 1], [1]) }
}



// wavetable buffer plotter

+ Buffer {
	plotWavetable { |name, bounds|
		this.getToFloatArray(wait: 0.05, action: { |v|
			{ v[0, 2..].plot(name, bounds) }.defer
		});
	}
	
	duration { ^numFrames / sampleRate }
}
